import logging
import re
from typing import Dict, Union

logger = logging.getLogger(__name__)


class FeatureDict(Dict[str, Union[str, bool]]):
    """文節，基本句，形態素の素性情報を表すクラス．"""

    PAT = re.compile(r"(?P<feats>(<[^>]+>)*)")
    IGNORE_TAG_PREFIXES = {"rel "}
    FEATURE_PAT = re.compile(rf"<(?!({'|'.join(IGNORE_TAG_PREFIXES)}))(?P<key>([^:\"]|\".*?\")+?)(:(?P<value>.+?))?>")

    def __setitem__(self, key, value) -> None:
        if key == "rel":
            logger.warning(
                f"Adding 'rel' to {self.__class__.__name__} is not supported and was ignored. Instead, add a RelTag "
                f"object to BasePhrase.rels and call Document.reparse()."
            )
            return
        super().__setitem__(key, value)

    @classmethod
    def from_fstring(cls, fstring: str) -> "FeatureDict":
        """素性文字列をパースして辞書型に変換する．
        e.g., "<正規化代表表記:遅れる/おくれる>" -> {"正規化代表表記": "遅れる/おくれる"}

        Args:
            fstring: KNP 形式における素性文字列．
        """
        features = {}
        for match in cls.FEATURE_PAT.finditer(fstring):
            features[match.group("key")] = match.group("value") or True
        return cls(features)

    def to_fstring(self) -> str:
        """素性文字列に変換．"""
        return "".join(self._item_to_fstring(k, v) for k, v in self.items())

    @staticmethod
    def _item_to_fstring(key: str, value: Union[str, bool]) -> str:
        if value is False:
            return ""
        if value is True:
            return f"<{key}>"
        return f"<{key}:{value}>"



from .dependency import DepType
from .feature import FeatureDict
from .named_entity import NamedEntity, NamedEntityCategory
from .semantics import SemanticsDict

__all__ = [
    "DepType",
    "FeatureDict",
    "SemanticsDict",
    "NamedEntity",
    "NamedEntityCategory",
]



from abc import ABC, abstractmethod
from multiprocessing import Pool
from typing import List, Sequence, Union

from rhoknp.units import Document, Sentence


class Processor(ABC):
    """解析器の基底クラス．"""

    @abstractmethod
    def apply_to_document(self, document: Union[Document, str]) -> Document:
        """文書に解析器を適用する．

        Args:
            document (Union[Document, str]): 文書．
        """
        raise NotImplementedError

    def batch_apply_to_documents(self, documents: Sequence[Union[Document, str]], processes: int = 0) -> List[Document]:
        """複数文書に解析器を適用する．

        Args:
            documents (Sequence[Union[Document, str]]): 文書のリスト．
            processes (int, optional): 並列処理数．0以下の場合はシングルプロセスで処理する．
        """
        if processes < 1:
            return list(map(self.apply_to_document, documents))
        with Pool(processes=processes) as pool:
            return pool.map(self.apply_to_document, documents)

    @abstractmethod
    def apply_to_sentence(self, sentence: Union[Sentence, str]) -> Sentence:
        """文に解析器を適用する．

        Args:
            sentence (Union[Sentence, str]): 文．
        """
        raise NotImplementedError

    def batch_apply_to_sentences(self, sentences: Sequence[Union[Sentence, str]], processes: int = 0) -> List[Sentence]:
        """複数文に解析器を適用する．

        Args:
            sentences (Sequence[Union[Sentence, str]]): 文のリスト．
            processes (int, optional): 並列処理数．0以下の場合はシングルプロセスで処理する．
        """
        if processes < 1:
            return list(map(self.apply, sentences))
        with Pool(processes=processes) as pool:
            return pool.map(self.apply, sentences)

    def apply(self, sentence: Union[Sentence, str]) -> Sentence:
        """文に解析器を適用する．

        Args:
            sentence (Union[Sentence, str]): 文．
        """
        return self.apply_to_sentence(sentence)

    def batch_apply(self, sentences: Sequence[Union[Sentence, str]], processes: int = 0) -> List[Sentence]:
        """複数文に解析器を適用する．

        Args:
            sentences (Sequence[Union[Sentence, str]]): 文のリスト．
            processes (int, optional): 並列処理数．0以下の場合はシングルプロセスで処理する．
        """
        return self.batch_apply_to_sentences(sentences, processes)



import logging
import re
from enum import Enum
from typing import Any, ClassVar, Optional

logger = logging.getLogger(__name__)


class ExophoraReferentType(Enum):
    WRITER = "著者"
    READER = "読者"
    UNSPECIFIED_PERSON = "不特定:人"
    UNSPECIFIED_MATTER = "不特定:物"
    UNSPECIFIED_SITUATION = "不特定:状況"
    PREVIOUS_SENTENCE = "前文"
    NEXT_SENTENCE = "後文"
    OTHER = "OTHER"


class ExophoraReferent:
    """外界照応における照応先を表すクラス．"""

    PAT: ClassVar[re.Pattern] = re.compile(
        rf"^(?P<type>{'|'.join(t.value for t in ExophoraReferentType if t != ExophoraReferentType.OTHER)})"
        rf"(?P<index>[０-９\d]*)$"
    )

    def __init__(self, text: str) -> None:
        self.index: Optional[int] = None
        self._other_text: Optional[str] = None
        match: Optional[re.Match[str]] = self.PAT.match(text)
        if match is None:
            logger.warning(f"unknown exophora referent found: {text}")
            self.type = ExophoraReferentType.OTHER
            self._other_text = text
        else:
            if index := match.group("index"):
                self.index = int(index)
            self.type = ExophoraReferentType(match.group("type"))

    @property
    def text(self) -> str:
        """外界照応の照応先を表すテキスト表現．"""
        if self.type != ExophoraReferentType.OTHER:
            return str(self.type.value) + str(self.index or "")
        else:
            assert self._other_text is not None
            return self._other_text

    @property
    def is_singleton(self) -> bool:
        """文書中に1つしか存在しないエンティティであれば True．"""
        if self.type in (ExophoraReferentType.WRITER, ExophoraReferentType.READER):
            return True
        if self.index is not None:
            return True
        return False

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={repr(self.text)})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, type(self)) is False or self.type != other.type:
            return False
        if self.type == ExophoraReferentType.OTHER:
            return self._other_text == other._other_text
        else:
            return self.index == other.index



import copy
import logging
import re
from collections import defaultdict
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from rhoknp.cohesion.argument import Argument, ArgumentType, EndophoraArgument, ExophoraArgument
from rhoknp.cohesion.exophora import ExophoraReferent
from rhoknp.cohesion.predicate import Predicate
from rhoknp.cohesion.rel import RelMode

if TYPE_CHECKING:
    from rhoknp.units.base_phrase import BasePhrase

logger = logging.getLogger(__name__)


class CaseInfoFormat(Enum):
    CASE = auto()  #: 格解析フォーマット
    PAS = auto()  #: 述語項構造フォーマット


class Pas:
    """述語項構造クラス．

    Args:
        predicate: 述語．
    """

    ARGUMENT_PAT = re.compile(r"([^/;]+/[CNODEU-]/[^/]+/(-?\d*)/(-?\d*)/[^/;]+)")  # matches for "ガ/N/彼/0/0/5"

    def __init__(self, predicate: Predicate):
        self._predicate = predicate
        predicate.pas = self
        self._arguments: Dict[str, List[Argument]] = defaultdict(list)
        self.modes: Dict[str, RelMode] = {}

    @property
    def predicate(self) -> Predicate:
        """述語．"""
        return self._predicate

    @property
    def cases(self) -> List[str]:
        """属する全ての項の持つ格を集めたリスト．"""
        return [case for case, args in self._arguments.items() if args]

    @property
    def sid(self) -> str:
        """文 ID．"""
        return self._predicate.sid

    @classmethod
    def from_pas_string(cls, base_phrase: "BasePhrase", fstring: str, format_: CaseInfoFormat) -> "Pas":
        """PAS 文字列から述語項構造を生成する．

        Args:
            base_phrase: 述語となる基本句．
            fstring: 述語項構造を表す素性文字列（e.g., "食べる/たべる:動2:ガ/C/太郎/0/0/1;ヲ/C/パン/1/0/1;ニ/U/-/-/-/-;デ/U/-/-/-/-;カラ/U/-/-/-/-;時間/U/-/-/-/-"）
            format_: fstring における述語項構造のフォーマット．
        """
        # language=RegExp
        cfid_pat = r"(.*?):([^:/]+?)"  # 食べる/たべる:動1
        match = re.match(
            r"{cfid}(:(?P<args>{args}(;{args})*))?$".format(cfid=cfid_pat, args=cls.ARGUMENT_PAT.pattern),
            fstring,
        )

        if match is None:
            logger.warning(f"invalid tag format: '{fstring}' is ignored")
            return cls(Predicate(base_phrase))

        cfid = match.group(1) + ":" + match.group(2)
        predicate = Predicate(unit=base_phrase, cfid=cfid)

        if match.group(3) is None:  # <述語項構造:束の間/つかのま:判0> など
            return cls(predicate)

        pas = cls(predicate)
        for match_arg in cls.ARGUMENT_PAT.finditer(match.group("args")):
            case, case_flag, surf, *fields = match_arg.group(0).split("/")
            if case_flag in ("U", "-"):
                continue
            arg_type = ArgumentType(case_flag)
            arg: Argument
            if format_ == CaseInfoFormat.CASE:
                tid, sdist, sid = int(fields[0]), int(fields[1]), fields[2]
                assert arg_type != ArgumentType.EXOPHORA
                if sdist == 0:
                    sentence = base_phrase.sentence
                else:
                    if (sentence_index := base_phrase.sentence.index - sdist) < 0:
                        logger.warning(f"sentence index out of range: {sentence_index} in {base_phrase.sentence.sid}")
                        continue
                    sentence = base_phrase.document.sentences[sentence_index]
                assert sentence.sid == sid
                arg_base_phrase = sentence.base_phrases[tid]
                assert surf in arg_base_phrase.text
                pas.add_argument(case, arg_base_phrase, arg_type=arg_type)
            else:
                assert format_ == CaseInfoFormat.PAS
                sdist, tid, eid = int(fields[0]), int(fields[1]), int(fields[2])
                if arg_type == ArgumentType.EXOPHORA:
                    pas.add_special_argument(case, surf, eid)
                else:
                    if sdist == 0:
                        sentence = base_phrase.sentence
                    else:
                        if (sentence_index := base_phrase.sentence.index - sdist) < 0:
                            logger.warning(
                                f"sentence index out of range: {sentence_index} in {base_phrase.sentence.sid}"
                            )
                            continue
                        sentence = base_phrase.document.sentences[sentence_index]
                    arg_base_phrase = sentence.base_phrases[tid]
                    assert surf in arg_base_phrase.text
                    pas.add_argument(case, arg_base_phrase)
        return pas

    def get_arguments(
        self,
        case: str,
        relax: bool = True,
        include_nonidentical: bool = False,
        include_optional: bool = False,
    ) -> List[Argument]:
        """与えられた格の全ての項を返す．

        Args:
            case: 格．
            relax: True なら 共参照関係で結ばれた項も含めて出力する．
            include_nonidentical: True なら nonidentical な項も含めて出力する．
            include_optional: True なら修飾的表現（「すぐに」など）も含めて出力する．

        References:
            格・省略・共参照タグ付けの基準 3.2.1 修飾的表現
        """
        args = self._arguments[case]
        if include_nonidentical is True:
            args += self._arguments[case + "≒"]
        if include_optional is False:
            args = [arg for arg in args if arg.optional is False]
        pas = copy.copy(self)
        pas._arguments = copy.copy(self._arguments)
        pas._arguments[case] = copy.copy(args)

        sentence = self.predicate.base_phrase.sentence
        if relax is True and sentence.parent_unit is not None:
            entity_manager = sentence.document.entity_manager
            for arg in args:
                if isinstance(arg, ExophoraArgument):
                    entities = {entity_manager[arg.eid]}
                else:
                    assert isinstance(arg, EndophoraArgument)
                    entities = arg.base_phrase.entities_all if include_nonidentical else arg.base_phrase.entities
                for entity in entities:
                    if entity.exophora_referent is not None:
                        pas.add_special_argument(case, entity.exophora_referent, entity.eid)
                    for mention in entity.mentions:
                        if isinstance(arg, EndophoraArgument) and mention == arg.base_phrase:
                            continue
                        pas.add_argument(case, mention)
        return pas._arguments[case]

    def get_all_arguments(
        self,
        relax: bool = True,
        include_nonidentical: bool = False,
        include_optional: bool = False,
    ) -> Dict[str, List[Argument]]:
        """この述語項構造が持つ全ての項を格を key とする辞書形式で返す．

        Args:
            relax: True なら 共参照関係で結ばれた項も含めて出力する．
            include_nonidentical: True なら nonidentical な項も含めて出力する．
            include_optional: True なら修飾的表現（「すぐに」など）も含めて出力する．
        """
        all_arguments: Dict[str, List[Argument]] = {}
        for case in self.cases:
            all_arguments[case] = self.get_arguments(
                case,
                relax=relax,
                include_nonidentical=include_nonidentical,
                include_optional=include_optional,
            )
        return all_arguments

    def add_argument(
        self,
        case: str,
        base_phrase: "BasePhrase",
        mode: Optional[RelMode] = None,
        arg_type: Optional[ArgumentType] = None,
    ) -> None:
        """述語項構造に項を追加．

        Args:
            case: 項が持つ格．
            base_phrase: 項となる基本句．
            mode: 関係のモード．
            arg_type: 述語と項の関係タイプ．
        """
        argument = EndophoraArgument(
            base_phrase,
            arg_type or self._get_arg_type(self.predicate, base_phrase, case),
        )
        argument.pas = self
        if mode is not None:
            self.modes[case] = mode
        if argument not in self._arguments[case]:
            self._arguments[case].append(argument)

    def add_special_argument(
        self, case: str, exophora_referent: Union[ExophoraReferent, str], eid: int, mode: Optional[RelMode] = None
    ) -> None:
        """述語項構造に外界照応に対応する項を追加．

        Args:
            case: 項が持つ格．
            exophora_referent: 外界照応における照応先．．
            eid: エンティティ ID．
            mode: 関係のモード．
        """
        if isinstance(exophora_referent, str):
            exophora_referent = ExophoraReferent(exophora_referent)
        special_argument = ExophoraArgument(exophora_referent, eid)
        special_argument.pas = self
        if mode is not None:
            self.modes[case] = mode
        if special_argument not in self._arguments[case]:
            self._arguments[case].append(special_argument)

    def set_arguments_optional(self, case: str) -> None:
        """与えられた格に属する項をすべて修飾的表現として登録．

        Args:
            case: 対象の格．
        """
        if not self._arguments[case]:
            logger.info(f"no preceding argument found in {self.sid}. 'なし' is ignored")
            return
        for arg in self._arguments[case]:
            arg.optional = True
            logger.info(f"marked {arg} as optional in {self.sid}")

    @staticmethod
    def _get_arg_type(predicate: Predicate, arg_base_phrase: "BasePhrase", case: str) -> ArgumentType:
        if arg_base_phrase in predicate.base_phrase.children:
            dep_case = arg_base_phrase.features.get("係", "")
            assert isinstance(dep_case, str)
            dep_case = dep_case.rstrip("格")
            if (case == dep_case) or (case == "判ガ" and dep_case == "ガ") or (case == "ノ？" and dep_case == "ノ"):
                return ArgumentType.CASE_EXPLICIT
            else:
                return ArgumentType.CASE_HIDDEN
        elif predicate.base_phrase.parent and predicate.base_phrase.parent == arg_base_phrase:
            return ArgumentType.CASE_HIDDEN
        else:
            return ArgumentType.OMISSION

    def __repr__(self) -> str:
        return f"<{self.__module__}.{self.__class__.__name__}: {repr(self.predicate.text)}>"



import logging
from functools import cached_property
from typing import TYPE_CHECKING, List, Optional

from rhoknp.cohesion.discourse import DiscourseRelation
from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.phrase import Phrase
from rhoknp.units.unit import Unit

if TYPE_CHECKING:
    from rhoknp.units.document import Document
    from rhoknp.units.sentence import Sentence

logger = logging.getLogger(__name__)


class Clause(Unit):
    """節クラス．"""

    count = 0

    def __init__(self) -> None:
        super().__init__()

        # parent unit
        self._sentence: Optional["Sentence"] = None

        # child units
        self._phrases: Optional[List[Phrase]] = None

        self.discourse_relations: List[DiscourseRelation] = []  #: 談話関係のリスト．

        self.index = self.count  #: 文内におけるインデックス．
        Clause.count += 1

    def __post_init__(self) -> None:
        super().__post_init__()

        # Find discourse relations.
        # TODO: Use forward/backward clause function
        self.discourse_relations = []
        for key in self.end.features:
            if key.startswith("節-機能"):
                if discourse_relation := DiscourseRelation.from_clause_function_fstring(key, modifier=self):
                    self.discourse_relations.append(discourse_relation)
        if values := self.end.features.get("談話関係", None):
            assert isinstance(values, str)
            for value in values.split(";"):
                if discourse_relation := DiscourseRelation.from_discourse_relation_fstring(value, modifier=self):
                    self.discourse_relations.append(discourse_relation)

    @cached_property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        if self.index > 0:
            return self.sentence.clauses[self.index - 1].global_index + 1
        if self.sentence.index == 0:
            return self.index
        return self.document.sentences[self.sentence.index - 1].clauses[-1].global_index + 1

    @property
    def parent_unit(self) -> Optional["Sentence"]:
        """上位の言語単位（文）．未登録なら None．"""
        return self._sentence

    @property
    def child_units(self) -> Optional[List[Phrase]]:
        """下位の言語単位（文節）．解析結果にアクセスできないなら None．"""
        return self._phrases

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.sentence.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._sentence is None:
            raise AttributeError("sentence has not been set")
        return self._sentence

    @sentence.setter
    def sentence(self, sentence: "Sentence") -> None:
        """文．

        Args:
            sentence: 文．
        """
        self._sentence = sentence

    @property
    def phrases(self) -> List[Phrase]:
        """文節のリスト．"""
        if self._phrases is None:
            raise AssertionError
        return self._phrases

    @phrases.setter
    def phrases(self, phrases: List[Phrase]) -> None:
        """文節のリスト．

        Args:
            phrases: 文節のリスト．
        """
        for phrase in phrases:
            phrase.clause = self
        self._phrases = phrases

    @property
    def base_phrases(self) -> List[BasePhrase]:
        """基本句のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [base_phrase for phrase in self.phrases for base_phrase in phrase.base_phrases]

    @property
    def morphemes(self) -> List[Morpheme]:
        """形態素のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [morpheme for base_phrase in self.base_phrases for morpheme in base_phrase.morphemes]

    @cached_property
    def head(self) -> BasePhrase:
        """節主辞の基本句．"""
        for base_phrase in self.base_phrases:
            if base_phrase.features and "節-主辞" in base_phrase.features:
                return base_phrase
        raise AssertionError

    @property
    def end(self) -> BasePhrase:
        """節区切の基本句．"""
        return self.base_phrases[-1]

    @cached_property
    def parent(self) -> Optional["Clause"]:
        """係り先の節．ないなら None．"""
        head_parent = self.head.parent
        while head_parent in self.base_phrases:
            head_parent = head_parent.parent
        for clause in self.sentence.clauses:
            if head_parent in clause.base_phrases:
                return clause
        return None

    @cached_property
    def children(self) -> List["Clause"]:
        """この節に係っている節のリスト．"""
        return [clause for clause in self.sentence.clauses if clause.parent == self]

    @cached_property
    def is_adnominal(self) -> bool:
        """連体修飾節なら True．"""
        return self.end.features.get("節-区切", "") == "連体修飾"

    @classmethod
    def from_knp(cls, knp_text: str) -> "Clause":
        """節クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
        """
        clause = cls()
        phrases = []
        phrase_lines: List[str] = []
        for line in knp_text.split("\n"):
            if not line.strip():
                continue
            if line.startswith("*") and phrase_lines:
                phrases.append(Phrase.from_knp("\n".join(phrase_lines)))
                phrase_lines = []
            phrase_lines.append(line)
        else:
            phrase = Phrase.from_knp("\n".join(phrase_lines))
            phrases.append(phrase)
        clause.phrases = phrases
        return clause

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        return "".join(phrase.to_knp() for phrase in self.phrases)



import logging
from typing import TYPE_CHECKING, List, Optional, Set

from rhoknp.cohesion.argument import ExophoraArgument
from rhoknp.cohesion.exophora import ExophoraReferent

if TYPE_CHECKING:
    from rhoknp.units.base_phrase import BasePhrase

logger = logging.getLogger(__name__)


class Entity:
    """共参照におけるエンティティ．

    Args:
        eid: エンティティ ID．
        exophora_referent: 自身が外界照応の照応先に対応するなら照応先の種類. 対応しないなら None.

    Attributes:
        eid: エンティティ ID
        exophora_referent: 外界照応の照応先．対応するものがなければ None.
        mentions: このエンティティを参照するメンションの集合．
        mentions_nonidentical: このエンティティを≒関係で参照するメンションの集合．
    """

    def __init__(self, eid: int, exophora_referent: Optional[ExophoraReferent] = None) -> None:
        self.eid: int = eid
        self.exophora_referent: Optional[ExophoraReferent] = exophora_referent
        self.mentions: Set["BasePhrase"] = set()
        self.mentions_nonidentical: Set["BasePhrase"] = set()

    @property
    def mentions_all(self) -> Set["BasePhrase"]:
        """nonidentical を含めたこのエンティティを参照する全てのメンションの集合．"""
        return self.mentions | self.mentions_nonidentical

    def add_mention(self, mention: "BasePhrase", nonidentical: bool = False) -> None:
        """このエンティティを参照するメンションを追加．

        When an identical mention is added and the mention has already been registered as a nonidentical
        mention, it will be overwritten as identical.

        Args:
            mention: 追加対象のメンション．
            nonidentical: メンションが nonidentical ("≒" 付きでアノテーションされている) なら True．
        """
        if nonidentical:
            if mention in self.mentions_all:
                return
            mention.entities_nonidentical.add(self)
            self.mentions_nonidentical.add(mention)
        else:
            if mention in self.mentions_nonidentical:
                self.remove_mention(mention)
            mention.entities.add(self)
            self.mentions.add(mention)

    def remove_mention(self, mention: "BasePhrase") -> None:
        """このエンティティを参照するメンションを削除．"""
        if mention in self.mentions:
            self.mentions.remove(mention)
            mention.entities.remove(self)
        if mention in self.mentions_nonidentical:
            self.mentions_nonidentical.remove(mention)
            mention.entities_nonidentical.remove(self)

    def __str__(self) -> str:
        if self.exophora_referent:
            return str(self.exophora_referent)
        if self.mentions:
            return list(self.mentions)[0].__str__()
        elif self.mentions_nonidentical:
            return list(self.mentions_nonidentical)[0].__str__()
        else:
            return ""

    def __repr__(self) -> str:
        items = [repr(self.eid)]
        items += [repr(m.text) for m in self.mentions]
        items += [repr(m.text) for m in self.mentions_nonidentical]
        if self.exophora_referent:
            items.append(repr(self.exophora_referent))
        return f"<{self.__module__}.{self.__class__.__name__}: {', '.join(items)}>"

    def __hash__(self):
        return hash(self.eid)


class EntityManager:
    """文書全体のエンティティを管理．

    Attributes:
        entities: エンティティのリスト．
    """

    def __init__(self):
        self.entities: List[Entity] = []

    def get_or_create_entity(
        self, exophora_referent: Optional[ExophoraReferent] = None, eid: Optional[int] = None
    ) -> Entity:
        """自身が参照するエンティティを作成．

        exophora_referent が singleton entity だった場合を除き、新しく Entity のインスタンスを作成して返す．
        singleton entity とは、「著者」や「不特定:人１」などの文書中に必ず一つしか存在しないような entity．
        一方で、「不特定:人」や「不特定:物」は複数存在しうるので singleton entity ではない．
        eid を指定しない場合、最後に作成した entity の次の eid を選択．

        Args:
            exophora_referent: 外界照応における照応先．対応するものがなければ None．
            eid: エンティティ ID．None の場合自動で割り振る．

        Returns:
             Entity: 作成されたエンティティ．
        """
        if exophora_referent is not None and exophora_referent.is_singleton is True:
            entities = [e for e in self.entities if exophora_referent == e.exophora_referent]
            # すでに singleton entity が存在した場合、新しい entity は作らずにその entity を返す
            if entities:
                assert len(entities) == 1  # singleton entity が1つしかないことを保証
                return entities[0]
        eids: List[int] = [e.eid for e in self.entities]
        if eid in eids:
            _eid = eid
            eid = max(eids) + 1
            logger.warning(f"{_eid} is already used. use eid: {eid} instead.")
        elif eid is None or eid < 0:
            eid = max(eids) + 1 if eids else 0
        entity = Entity(eid, exophora_referent=exophora_referent)
        self.entities.append(entity)
        return entity

    def merge_entities(
        self,
        source_mention: "BasePhrase",
        target_mention: Optional["BasePhrase"],
        source_entity: Entity,
        target_entity: Entity,
        nonidentical: bool,
    ) -> None:
        """2つのエンティティをマージ．

        source_mention と source_entity, target_mention と target_entity の間には mention が張られているが、
        source と target 間には張られていないので、add_mention する．
        source_entity と target_entity が同一のエンティティであり、exophor も同じか片方が None ならば target_entity の方を削除する．

        Args:
            source_mention: ソース側メンション．
            target_mention: ターゲット側メンション．メンションが存在しない場合は None．
            source_entity: ソース側エンティティ．
            target_entity: ターゲット側エンティティ．
            nonidentical: ソース側メンションとターゲット側メンションの関係が nonidentical なら True．
        """
        nonidentical_tgt = (target_mention is not None) and target_mention.is_nonidentical_to(target_entity)
        if source_entity not in source_mention.entities_all:
            return
        nonidentical_src = source_mention.is_nonidentical_to(source_entity)
        if source_entity is target_entity:
            if not nonidentical:
                # source_entity (target_entity), source_mention, target_mention の三角形のうち2辺が identical ならもう1辺も identical
                if (not nonidentical_src) and nonidentical_tgt:
                    assert target_mention is not None
                    source_entity.add_mention(target_mention, nonidentical=False)
                if nonidentical_src and (not nonidentical_tgt):
                    source_entity.add_mention(source_mention, nonidentical=False)
            return
        if target_mention is not None:
            source_entity.add_mention(target_mention, nonidentical=(nonidentical or nonidentical_src))
        target_entity.add_mention(source_mention, nonidentical=(nonidentical or nonidentical_tgt))
        # source_entity と target_entity が同一でない可能性が捨てきれない場合、target_entity は削除しない
        if nonidentical_src or nonidentical or nonidentical_tgt:
            return
        # source_entity と target_entity が同一でも exophor が異なれば target_entity は削除しない
        if (
            source_entity.exophora_referent is not None
            and target_entity.exophora_referent is not None
            and source_entity.exophora_referent != target_entity.exophora_referent
        ):
            return
        # 以下 target_entity を削除する準備
        if source_entity.exophora_referent is None:
            source_entity.exophora_referent = target_entity.exophora_referent
        for tm in target_entity.mentions_all:
            source_entity.add_mention(tm, nonidentical=tm.is_nonidentical_to(target_entity))
        # argument も eid を持っているので eid が変わった場合はこちらも更新
        pas_list = source_mention.document.pas_list()
        for arg in [arg for pas in pas_list for args in pas.get_all_arguments(relax=False).values() for arg in args]:
            if isinstance(arg, ExophoraArgument) and arg.eid == target_entity.eid:
                arg.eid = source_entity.eid
        self.delete_entity(target_entity)  # delete target entity

    def delete_entity(self, entity: Entity) -> None:
        """エンティティを削除．

        Remove the target entity from all the mentions of the entity as well as from self.entities.
        Note that entity IDs can have a missing number.

        Args:
            entity: 削除対象のエンティティ．
        """
        if entity not in self.entities:
            return
        for mention in entity.mentions_all:
            entity.remove_mention(mention)
        self.entities.remove(entity)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(entities={repr(self.entities)})"

    def __getitem__(self, eid: int) -> Entity:
        es = [e for e in self.entities if e.eid == eid]
        if len(es) == 0:
            raise KeyError(f"entity ID: {eid} not found.")
        return es[0]



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from rhoknp.cohesion.pas import Pas
    from rhoknp.units.base_phrase import BasePhrase


class Predicate:
    def __init__(self, unit: "BasePhrase", cfid: Optional[str] = None):
        self.unit: "BasePhrase" = unit
        self.cfid: Optional[str] = cfid
        self._pas: Optional["Pas"] = None

    @property
    def base_phrase(self) -> "BasePhrase":
        """基本句．"""
        return self.unit

    @property
    def text(self) -> str:
        """表層文字列．"""
        return self.unit.text

    @property
    def sid(self) -> str:
        """文 ID．"""
        if self.unit.sentence is None:
            raise AttributeError("sentence has not been set")
        if self.unit.sentence.sid is None:
            raise AttributeError(f"sid of sentence: {repr(self.unit.sentence)} has not been set")
        return self.unit.sentence.sid

    @property
    def pas(self) -> "Pas":
        """述語項構造．"""
        if self._pas is None:
            raise AttributeError("pas has not been set")
        return self._pas

    @pas.setter
    def pas(self, pas: "Pas") -> None:
        """述語項構造．

        Args:
            pas: 述語項構造．
        """
        self._pas = pas

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"<{self.__module__}.{self.__class__.__name__}: {repr(self.text)}>"



import logging
import re
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, ClassVar, List, Optional

if TYPE_CHECKING:
    from rhoknp.units.morpheme import Morpheme

logger = logging.getLogger(__name__)


class NamedEntityCategory(Enum):
    """固有表現カテゴリを表す列挙体．"""

    ORGANIZATION = "ORGANIZATION"
    PERSON = "PERSON"
    LOCATION = "LOCATION"
    ARTIFACT = "ARTIFACT"
    DATE = "DATE"
    TIME = "TIME"
    MONEY = "MONEY"
    PERCENT = "PERCENT"
    OPTIONAL = "OPTIONAL"

    @classmethod
    def has_value(cls, value: str) -> bool:
        return any(value == item.value for item in cls)


@dataclass
class NamedEntity:
    """固有表現を表すクラス．"""

    PAT: ClassVar[re.Pattern] = re.compile(r"(?P<cat>\w+):(?P<name>[^>]+)")

    category: NamedEntityCategory
    morphemes: List["Morpheme"]

    @property
    def text(self) -> str:
        return "".join(m.text for m in self.morphemes)

    def __str__(self) -> str:
        return self.text

    @classmethod
    def from_fstring(cls, fstring: str, candidate_morphemes: List["Morpheme"]) -> Optional["NamedEntity"]:
        """KNP における素性文字列からオブジェクトを作成．"""
        match = cls.PAT.match(fstring)
        if match is None:
            logger.warning(f"{fstring} is not a valid NE fstring")
            return None
        category = match.group("cat")
        if not NamedEntityCategory.has_value(category):
            logger.warning(f"{candidate_morphemes[0].sentence.sid}: unknown NE category: {category}")
            return None
        name = match.group("name")
        if (span := cls._find_morpheme_span(name, candidate_morphemes)) is None:
            logger.warning(f"{candidate_morphemes[0].sentence.sid}: morpheme span of '{name}' not found")
            return None
        return NamedEntity(NamedEntityCategory(category), candidate_morphemes[span.start : span.stop])

    def to_fstring(self) -> str:
        """素性文字列に変換．"""
        return f"<NE:{self.category.value}:{self.text}>"

    @staticmethod
    def _find_morpheme_span(name: str, candidates: List["Morpheme"]) -> Optional[range]:
        """name にマッチする形態素の範囲を返す．

        Args:
            name: 固有表現の文字列
            candidates: 固有表現を構成する候補形態素のリスト
        """
        stop = len(candidates)
        while stop > 0:
            for start in reversed(range(stop)):
                if "".join(m.text for m in candidates[start:stop]) == name:
                    return range(start, stop)
            stop -= 1
        return None



import logging
import re
from functools import cached_property
from typing import TYPE_CHECKING, List, Optional, Set

from rhoknp.cohesion.coreference import Entity, EntityManager
from rhoknp.cohesion.exophora import ExophoraReferent
from rhoknp.cohesion.pas import CaseInfoFormat, Pas
from rhoknp.cohesion.predicate import Predicate
from rhoknp.cohesion.rel import RelMode, RelTag, RelTagList
from rhoknp.props.dependency import DepType
from rhoknp.props.feature import FeatureDict
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.unit import Unit
from rhoknp.utils.constants import ALL_CASES, ALL_COREFS

if TYPE_CHECKING:
    from rhoknp.units.clause import Clause
    from rhoknp.units.document import Document
    from rhoknp.units.phrase import Phrase
    from rhoknp.units.sentence import Sentence

logger = logging.getLogger(__name__)


class BasePhrase(Unit):
    """基本句クラス．"""

    KNP_PAT = re.compile(
        rf"^\+( (?P<pid>-1|\d+)(?P<dtype>[{''.join(e.value for e in DepType)}]))?( (?P<tags>(<[^>]+>)*))?$"
    )
    count = 0

    def __init__(
        self,
        parent_index: Optional[int],
        dep_type: Optional[DepType],
        features: Optional[FeatureDict] = None,
        rels: Optional[RelTagList] = None,
    ):
        super().__init__()

        # parent unit
        self._phrase: Optional["Phrase"] = None

        # child units
        self._morphemes: Optional[List[Morpheme]] = None

        self.parent_index: Optional[int] = parent_index  #: 係り先の基本句の文内におけるインデックス．
        self.dep_type: Optional[DepType] = dep_type  #: 係り受けの種類．
        self.features: FeatureDict = features or FeatureDict()  #: 素性．
        self.rels: RelTagList = rels or RelTagList()  #: 基本句間関係．
        self.pas: Optional["Pas"] = None  #: 述語項構造．
        self.entities: Set[Entity] = set()  #: 参照しているエンティティ．
        self.entities_nonidentical: Set[Entity] = set()  #: ≒で参照しているエンティティ．

        self.index = self.count  #: 文内におけるインデックス．
        BasePhrase.count += 1

    def __post_init__(self) -> None:
        super().__post_init__()

        # Parse the PAS tag.
        if "述語項構造" in self.features:
            pas_string = self.features["述語項構造"]
            assert isinstance(pas_string, str)
            pas = Pas.from_pas_string(self, pas_string, format_=CaseInfoFormat.PAS)
        elif "格解析結果" in self.features:
            pas_string = self.features["格解析結果"]
            assert isinstance(pas_string, str)
            pas = Pas.from_pas_string(self, pas_string, format_=CaseInfoFormat.CASE)
        else:
            pas = Pas(Predicate(self))
        self.pas = pas

        # Parse the rel tag if this unit is a piece of a document.
        if self.sentence.has_document is False:
            return
        for rel in self.rels:
            if rel.sid == "":
                rel.sid = self.sentence.sid
            if rel.type in ALL_CASES:
                self._add_pas(rel)
            elif rel.type in ALL_COREFS and rel.mode in (None, RelMode.AND):  # ignore "OR" and "?"
                self._add_coreference(rel)

    @cached_property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        if self.index > 0:
            return self.sentence.base_phrases[self.index - 1].global_index + 1
        if self.sentence.index == 0:
            return self.index
        return self.document.sentences[self.sentence.index - 1].base_phrases[-1].global_index + 1

    @property
    def parent_unit(self) -> Optional["Phrase"]:
        """上位の言語単位（文節）．未登録なら None．"""
        return self._phrase

    @property
    def child_units(self) -> Optional[List[Morpheme]]:
        """下位の言語単位（形態素）．解析結果にアクセスできないなら None．"""
        return self._morphemes

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.sentence

    @property
    def clause(self) -> "Clause":
        """節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.clause

    @property
    def phrase(self) -> "Phrase":
        """文節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._phrase is None:
            raise AttributeError("phrase has not been set")
        return self._phrase

    @phrase.setter
    def phrase(self, phrase: "Phrase") -> None:
        """文節．

        Args:
            phrase: 文節．
        """
        self._phrase = phrase

    @property
    def morphemes(self) -> List[Morpheme]:
        """形態素のリスト．"""
        if self._morphemes is None:
            raise AssertionError
        return self._morphemes

    @morphemes.setter
    def morphemes(self, morphemes: List[Morpheme]) -> None:
        """形態素のリスト．

        Args:
            morphemes: 形態素のリスト．
        """
        for morpheme in morphemes:
            morpheme.base_phrase = self
        self._morphemes = morphemes

    @property
    def head(self) -> Morpheme:
        """主辞の形態素．"""
        feature_to_priority = {"内容語": 0, "準内容語": 1, "基本句-主辞": 2}
        head = self.morphemes[0]
        current_priority = -1
        for morpheme in self.morphemes:
            if not morpheme.features:
                continue
            for feature, priority in feature_to_priority.items():
                if feature in morpheme.features and priority > current_priority:
                    head = morpheme
                    current_priority = priority
        return head

    @property
    def parent(self) -> Optional["BasePhrase"]:
        """係り先の基本句．ないなら None．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self.parent_index is None:
            raise AttributeError("parent_index has not been set")
        if self.parent_index == -1:
            return None
        return self.sentence.base_phrases[self.parent_index]

    @cached_property
    def children(self) -> List["BasePhrase"]:
        """この基本句に係っている基本句のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [base_phrase for base_phrase in self.sentence.base_phrases if base_phrase.parent == self]

    @property
    def entities_all(self) -> Set[Entity]:
        """nonidentical も含めた参照している全エンティティの集合．"""
        return self.entities | self.entities_nonidentical

    @classmethod
    def from_knp(cls, knp_text: str) -> "BasePhrase":
        """基本句クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
        """
        first_line, *lines = knp_text.split("\n")
        match = cls.KNP_PAT.match(first_line)
        if match is None:
            raise ValueError(f"malformed line: {first_line}")
        parent_index = int(match.group("pid")) if match.group("pid") is not None else None
        dep_type = DepType(match.group("dtype")) if match.group("dtype") is not None else None
        features = FeatureDict.from_fstring(match.group("tags") or "")
        rels = RelTagList.from_fstring(match.group("tags") or "")
        base_phrase = cls(parent_index, dep_type, features, rels)

        morphemes: List[Morpheme] = []
        for line in lines:
            if not line.strip():
                continue
            morphemes.append(Morpheme.from_jumanpp(line))
        base_phrase.morphemes = morphemes
        return base_phrase

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        ret = "+"
        if self.parent_index is not None:
            assert self.dep_type is not None
            ret += f" {self.parent_index}{self.dep_type.value}"
        if self.rels or self.features:
            ret += " "
            ret += self.rels.to_fstring()
            ret += self.features.to_fstring()
        ret += "\n"
        ret += "".join(morpheme.to_knp() for morpheme in self.morphemes)
        return ret

    def get_coreferents(self, include_nonidentical: bool = False, include_self: bool = False) -> Set["BasePhrase"]:
        """この基本句と共参照している基本句の集合を返却．

        Args:
            include_nonidentical: nonidentical なメンションを含めるなら True．
            include_self: 自身を含めるなら True．

        Returns:
            共参照している基本句の集合．
        """
        mentions: Set["BasePhrase"] = set()
        for entity in self.entities:
            mentions.update(entity.mentions)
        if include_nonidentical is True:
            for entity in self.entities_nonidentical:
                mentions.update(entity.mentions)
        if include_self is False and self in mentions:
            mentions.remove(self)
        return mentions

    def is_nonidentical_to(self, entity: Entity) -> bool:
        """エンティティに対して自身が nonidentical な場合に True を返す．

        Raises:
            AssertionError: 自身が参照しないエンティティだった場合．
        """
        if entity in self.entities:
            return False
        else:
            assert entity in self.entities_nonidentical, f"non-referring entity: {entity}"
            return True

    def add_entity(self, entity: Entity, nonidentical: bool = False) -> None:
        """エンティティを追加．"""
        if nonidentical:
            self.entities_nonidentical.add(entity)
        else:
            self.entities.add(entity)
        entity.add_mention(self, nonidentical=nonidentical)

    def _add_pas(self, rel: RelTag) -> None:
        """述語項構造を追加．"""
        entity_manager: EntityManager = self.document.entity_manager
        assert self.pas is not None
        if rel.sid is not None:
            if (arg_base_phrase := self._get_target_base_phrase(rel)) is None:
                return
            if not arg_base_phrase.entities:
                arg_base_phrase.add_entity(entity_manager.get_or_create_entity())
            self.pas.add_argument(rel.type, arg_base_phrase, mode=rel.mode)
        else:
            if rel.target == "なし":
                self.pas.set_arguments_optional(rel.type)
                return
            # exophora
            entity = entity_manager.get_or_create_entity(ExophoraReferent(rel.target))
            self.pas.add_special_argument(rel.type, rel.target, eid=entity.eid, mode=rel.mode)

    def _add_coreference(self, rel: RelTag) -> None:
        """共参照関係を追加．"""
        entity_manager: EntityManager = self.document.entity_manager
        # create source entity
        if not self.entities:
            self.add_entity(entity_manager.get_or_create_entity())

        nonidentical: bool = rel.type.endswith("≒")
        if rel.sid is not None:
            if (target_base_phrase := self._get_target_base_phrase(rel)) is None:
                return
            if target_base_phrase == self:
                logger.warning(f"{self.sentence.sid}: coreference with self found: {self}")
                return
            # create target entity
            if not target_base_phrase.entities:
                target_base_phrase.add_entity(entity_manager.get_or_create_entity())
            for source_entity in self.entities_all:
                for target_entity in target_base_phrase.entities_all:
                    entity_manager.merge_entities(self, target_base_phrase, source_entity, target_entity, nonidentical)
        else:
            # exophora
            for source_entity in self.entities_all:
                target_entity = entity_manager.get_or_create_entity(exophora_referent=ExophoraReferent(rel.target))
                entity_manager.merge_entities(self, None, source_entity, target_entity, nonidentical)

    def _get_target_base_phrase(self, rel: RelTag) -> Optional["BasePhrase"]:
        """rel が指す基本句を取得．"""
        sentences = [sent for sent in self.document.sentences if sent.sid == rel.sid]
        if not sentences:
            logger.warning(f"{self.sentence.sid}: relation with unknown sid found: {rel.sid}")
            return None
        sentence = sentences[0]
        assert rel.base_phrase_index is not None
        if rel.base_phrase_index >= len(sentence.base_phrases):
            logger.warning(f"{self.sentence.sid}: index out of range")
            return None
        target_base_phrase = sentence.base_phrases[rel.base_phrase_index]
        if not (set(rel.target) <= set(target_base_phrase.text)):
            logger.info(f"{self.sentence.sid}: rel target mismatch; '{rel.target}' vs '{target_base_phrase.text}'")
        return target_base_phrase

    def __hash__(self) -> int:
        return hash((self.global_index, self.sentence.sid))



from .argument import Argument, ArgumentType, EndophoraArgument, ExophoraArgument
from .coreference import Entity, EntityManager
from .discourse import DiscourseRelation, DiscourseRelationLabel, DiscourseRelationTag
from .exophora import ExophoraReferent, ExophoraReferentType
from .pas import Pas
from .predicate import Predicate
from .rel import RelMode, RelTag, RelTagList

__all__ = [
    "Pas",
    "Predicate",
    "Argument",
    "EndophoraArgument",
    "ExophoraArgument",
    "ArgumentType",
    "ExophoraReferent",
    "ExophoraReferentType",
    "Entity",
    "EntityManager",
    "RelTag",
    "RelTagList",
    "RelMode",
    "DiscourseRelation",
    "DiscourseRelationLabel",
    "DiscourseRelationTag",
]



from typing import Any, Dict

from rhoknp import Document
from rhoknp.utils.constants import ALL_CASES, ALL_COREFS


def get_document_statistics(document: Document) -> Dict[str, Dict[str, int]]:
    """文書の統計情報を取得．

    Args:
        document (Document): 文書．

    Returns:
        Dict[str, Dict[str, int]]: 統計情報．
    """
    stats: Dict[str, Any] = {"unit": {}, "cohesion": {}, "other": {}}
    # Unit
    if document.need_senter is False:
        stats["unit"]["sentence"] = len(document.sentences)
    if document.need_clause_tag is False:
        stats["unit"]["clause"] = len(document.clauses)
    if document.need_knp is False:
        stats["unit"]["phrase"] = len(document.phrases)
        stats["unit"]["base_phrase"] = len(document.base_phrases)
    if document.need_jumanpp is False:
        stats["unit"]["morpheme"] = len(document.morphemes)
    # Cohesion
    if document.need_knp is False:
        stats["cohesion"]["predicate"] = sum(
            len([rel for rel in bp.rels if rel.type in ALL_CASES]) > 0 for bp in document.base_phrases
        )
        stats["cohesion"]["argument"] = sum(
            len([rel for rel in bp.rels if rel.type in ALL_CASES]) for bp in document.base_phrases
        )
        stats["cohesion"]["coreference"] = sum(
            len([rel for rel in bp.rels if rel.type in ALL_COREFS]) for bp in document.base_phrases
        )
    if document.need_clause_tag is False:
        stats["cohesion"]["discourse"] = sum(len(clause.discourse_relations) for clause in document.clauses)
    if document.need_senter is False:
        stats["other"]["named_entity"] = sum(len(sentence.named_entities) for sentence in document.sentences)
    return stats



from .base_phrase import BasePhrase
from .clause import Clause
from .document import Document
from .morpheme import Morpheme
from .phrase import Phrase
from .sentence import Sentence

__all__ = ["Document", "Sentence", "Clause", "Phrase", "BasePhrase", "Morpheme"]



import json
from pathlib import Path

import typer
import yaml

from rhoknp import Document, __version__
from rhoknp.utils.draw_tree import draw_tree
from rhoknp.utils.stats import get_document_statistics

app = typer.Typer(help="rhoknp CLI utilities.")


def version_callback(value: bool):
    if value:
        typer.echo(f"rhoknp version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    _: bool = typer.Option(False, "--version", "-v", callback=version_callback, help="Show version and exit."),
):
    pass


@app.command(help="Print given file content in tree format.")
def show(
    knp_path: Path = typer.Argument(..., exists=True, dir_okay=False, help="Path to knp file to show"),
):
    doc = Document.from_knp(knp_path.read_text())
    for sent in doc.sentences:
        print(sent.comment)
        draw_tree(sent.base_phrases, show_pos=False)


@app.command(help="Show statistics of given KNP file.")
def stats(
    knp_path: Path = typer.Argument(
        ..., exists=True, dir_okay=False, help="Path to knp file to calculate statistics on."
    ),
    use_json: bool = typer.Option(False, "--json", "-j", help="Output statistics in JSON format."),
):
    doc = Document.from_knp(knp_path.read_text())
    doc_stats = get_document_statistics(doc)
    if use_json:
        typer.echo(json.dumps(doc_stats, ensure_ascii=False, indent=4))
    else:
        typer.echo(yaml.dump(doc_stats, allow_unicode=True, sort_keys=False))


if __name__ == "__main__":
    app()



import re
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, List, Optional


class RelMode(Enum):
    """同一の基本句に同一タイプの関係タグが複数付いている場合にそれらの関係を表す列挙体．

    .. note::
        * AND: 関係の対象が並列である．
            (例) 太郎と花子が学校から<帰った> (ガ格:太郎, ガ格:花子 [and])
        * OR: 「AかB」のように意味的に or である．
            (例) 私は田園調布か国立に<住みたい> (ガ格:私, ニ格:田園調布, ニ格:国立 [or])
        * AMBIGUOUS:いずれの解釈も妥当であり，文脈から判断ができない．
            (例) 高知県の橋本知事は…国籍条項を<撤廃する>方針を明らかにした (ガ格:高知県, ガ格:橋本知事 [？], ガ格:不特定:人 [？], ヲ格:条項, 外の関係:方針)

    .. note::
        target が「なし」の場合、同じタイプの関係タグが任意的要素であることを示す．
            (例) 太郎は一人で<立っていた> (ガ格:太郎, デ格:一人, デ格:なし [？])
    """

    AND = "AND"
    OR = "OR"
    AMBIGUOUS = "？"


@dataclass
class RelTag:
    """関係タグ付きコーパスにおける <rel> タグを表すクラス．"""

    PAT: ClassVar[re.Pattern] = re.compile(
        r'<rel type="(?P<type>\S+?)"( mode="(?P<mode>[^>]+?)")? target="(?P<target>.+?)"( sid="(?P<sid>.*?)" '
        r'id="(?P<id>\d+?)")?/>'
    )
    type: str
    target: str
    sid: Optional[str]
    base_phrase_index: Optional[int]
    mode: Optional[RelMode]

    def to_fstring(self) -> str:
        """素性文字列に変換．"""
        ret = f'<rel type="{self.type}"'
        if self.mode is not None:
            ret += f' mode="{self.mode.value}"'
        ret += f' target="{self.target}"'
        if self.sid is not None:
            assert self.base_phrase_index is not None
            ret += f' sid="{self.sid}" id="{self.base_phrase_index}"'
        ret += "/>"
        return ret


class RelTagList(List[RelTag]):
    """関係タグ付きコーパスにおける <rel> タグの列を表すクラス．"""

    @classmethod
    def from_fstring(cls, fstring: str) -> "RelTagList":
        """KNP における素性文字列からオブジェクトを作成．"""
        rels = []
        for match in RelTag.PAT.finditer(fstring):
            rels.append(
                RelTag(
                    type=match["type"],
                    target=match["target"],
                    sid=match["sid"],
                    base_phrase_index=int(match["id"]) if match["id"] else None,
                    mode=RelMode(match["mode"]) if match["mode"] else None,
                )
            )
        return cls(rels)

    def to_fstring(self) -> str:
        """素性文字列に変換．"""
        return "".join(rel.to_fstring() for rel in self)

    def __str__(self) -> str:
        return self.to_fstring()



from abc import ABC, abstractmethod
from typing import Any, Optional, Sequence


class Unit(ABC):
    def __init__(self) -> None:
        self.index: int = -1
        self._text: Optional[str] = None

    def __post_init__(self) -> None:
        if self.child_units is not None:
            for child_unit in self.child_units:
                child_unit.__post_init__()

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"<{self.__module__}.{self.__class__.__name__}: {repr(self.index)}, {repr(self.text)}>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, type(self)) is False:
            return False
        if self.parent_unit != other.parent_unit:
            return False
        return self.index == other.index

    @property
    def parent_unit(self) -> Optional["Unit"]:
        raise NotImplementedError

    @property
    @abstractmethod
    def child_units(self) -> Optional[Sequence["Unit"]]:
        raise NotImplementedError

    @property
    def text(self) -> str:
        if self._text is not None:
            return self._text
        elif self.child_units is not None:
            self._text = "".join(str(child_unit) for child_unit in self.child_units)
            return self._text
        raise AssertionError

    @text.setter
    def text(self, text: str) -> None:
        self._text = text



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rhoknp',
 'rhoknp.cohesion',
 'rhoknp.processors',
 'rhoknp.props',
 'rhoknp.units',
 'rhoknp.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['rhoknp = rhoknp.cli:app']}

setup_kwargs = {
    'name': 'rhoknp',
    'version': '0.4.1',
    'description': 'Yet another Python binding for Juman++/KNP',
    'long_description': '# rhoknp: Yet another Python binding for Juman++/KNP\n\n[![Test](https://img.shields.io/github/workflow/status/ku-nlp/rhoknp/test?logo=github&label=test&style=flat-square)](https://github.com/ku-nlp/rhoknp/actions/workflows/test.yml)\n[![Codecov](https://img.shields.io/codecov/c/github/ku-nlp/rhoknp?logo=codecov&style=flat-square)](https://codecov.io/gh/ku-nlp/rhoknp)\n[![PyPI](https://img.shields.io/pypi/v/rhoknp?style=flat-square)](https://pypi.org/project/rhoknp/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/rhoknp?style=flat-square)\n[![Documentation](https://img.shields.io/readthedocs/rhoknp?style=flat-square)](https://rhoknp.readthedocs.io/en/latest/?badge=latest)\n[![Code style - black](https://img.shields.io/badge/code%20style-black-222222?style=flat-square)](https://github.com/psf/black)\n\n*rhoknp* is a Python binding for [Juman++](https://github.com/ku-nlp/jumanpp) and [KNP](https://github.com/ku-nlp/knp).\n\n```python\nimport rhoknp\n\n# Perform language analysis by Juman++\njumanpp = rhoknp.Jumanpp()\nsentence = jumanpp.apply("電気抵抗率は電気の通しにくさを表す物性値である。")\n\n# Save language analysis by Juman++\nwith open("result.jumanpp", "wt") as f:\n    f.write(sentence.to_jumanpp())\n\n# Load language analysis by Juman++\nwith open("result.jumanpp", "rt") as f:\n    sentence = rhoknp.Sentence.from_jumanpp(f.read())\n\n# Perform language analysis by KNP\nknp = rhoknp.KNP()\nsentence = knp.apply(sentence)  # or knp.apply("電気抵抗率は...")\n\n# Save language analysis by KNP\nwith open("result.knp", "wt") as f:\n    f.write(sentence.to_knp())\n\n# Load language analysis by KNP\nwith open("result.knp", "rt") as f:\n    sentence = rhoknp.Sentence.from_knp(f.read())\n```\n\n## Requirements\n\n- Python 3.8+\n- [Juman++](https://github.com/ku-nlp/jumanpp) v2.0.0-rc3+\n- [KNP](https://github.com/ku-nlp/knp) 5.0+\n\n## Installation\n\n```shell\npip install rhoknp\n```\n\n## Documentation\n\n[https://rhoknp.readthedocs.io/en/latest/](https://rhoknp.readthedocs.io/en/latest/)\n\n## Quick tour\n\n*rhoknp* provides APIs to perform language analysis by Juman++ and KNP.\n\n```python\n# Perform language analysis by Juman++\njumanpp = rhoknp.Jumanpp()\nsentence = jumanpp.apply("電気抵抗率は電気の通しにくさを表す物性値である。")\n\n# Perform language analysis by KNP\nknp = rhoknp.KNP()\nsentence = knp.apply(sentence)  # or knp.apply("電気抵抗率は...")\n```\n\nSentence objects can be saved in the Juman/KNP format\n\n```python\n# Save language analysis by Juman++\nwith open("result.jumanpp", "wt") as f:\n    f.write(sentence.to_jumanpp())\n\n# Save language analysis by KNP\nwith open("result.knp", "wt") as f:\n    f.write(sentence.to_knp())\n```\n\nand recovered from Juman/KNP-format text.\n\n```python\n# Load language analysis by Juman++\nwith open("result.jumanpp", "rt") as f:\n    sentence = rhoknp.Sentence.from_jumanpp(f.read())\n\n# Perform language analysis by KNP\nwith open("result.knp", "rt") as f:\n    sentence = rhoknp.Sentence.from_knp(f.read())\n```\n\nIt is easy to access the linguistic units that make up a sentence.\n\n```python\nfor clause in sentence.clauses:\n    ...\nfor phrase in sentence.phrases:  # a.k.a. bunsetsu\n    ...\nfor base_phrase in sentence.base_phrases:  # a.k.a. kihon-ku\n    ...\nfor morpheme in sentence.morphemes:\n    ...\n```\n\n*rhoknp* also provides APIs for document-level language analysis.\n\n```python\ndocument = rhoknp.Document.from_raw_text(\n    "電気抵抗率は電気の通しにくさを表す物性値である。単に抵抗率とも呼ばれる。"\n)\n# If you know sentence boundaries, you can use `Document.from_sentences` instead.\ndocument = rhoknp.Document.from_sentences(\n    [\n        "電気抵抗率は電気の通しにくさを表す物性値である。",\n        "単に抵抗率とも呼ばれる。",\n    ]\n)\n```\n\nDocument objects can be handled in almost the same way as Sentence objects.\n\n```python\n# Perform language analysis by Juman++/KNP\ndocument = jumanpp.apply_to_document(document)\ndocument = knp.apply_to_document(document)\n\n# Save language analysis by Juman++/KNP\nwith open("result.jumanpp", "wt") as f:\n    f.write(document.to_jumanpp())\nwith open("result.knp", "wt") as f:\n    f.write(document.to_knp())\n\n# Load language analysis by Juman++/KNP\nwith open("result.jumanpp", "rt") as f:\n    document = rhoknp.Document.from_jumanpp(f.read())\nwith open("result.knp", "rt") as f:\n    document = rhoknp.Document.from_knp(f.read())\n\n# Access language units in the document\nfor sentence in document.sentences:\n    ...\nfor clause in document.clauses:\n    ...\nfor phrase in document.phrases:\n    ...\nfor base_phrase in document.base_phrases:\n    ...\nfor morpheme in document.morphemes:\n    ...\n```\n\nFor more information, explore the [examples](./examples) and [documentation](https://rhoknp.readthedocs.io/en/latest/).\n\n## Main differences from [pyknp](https://github.com/ku-nlp/pyknp/)\n\n- **Support document-level language analysis**: *rhoknp* can load and instantiate the result of document-level language analysis: i.e., cohesion analysis and discourse relation analysis.\n- **Strictly type-aware**: *rhoknp* is thoroughly annotated with type annotations. Efficient development is possible with the help of an IDE.\n- **Extensive test suite**: *rhoknp* is tested with an extensive test suite. See the code coverage at [Codecov](https://app.codecov.io/gh/ku-nlp/rhoknp).\n- **Support Python3.8+ only**\n\n## Reference\n\n- [KNP FORMAT](http://cr.fvcrc.i.nagoya-u.ac.jp/~sasano/knp/format.html)\n- [KNP - KUROHASHI-CHU-MURAWAKI LAB](https://nlp.ist.i.kyoto-u.ac.jp/?KNP)\n',
    'author': 'Hirokazu Kiyomaru',
    'author_email': 'h.kiyomaru@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ku-nlp/rhoknp',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging
import re
from typing import Callable, Iterator, List, Literal, Optional, TextIO, Union

from rhoknp import Sentence

logger = logging.getLogger(__name__)


def chunk_by_sentence(f: TextIO) -> Iterator[str]:
    """ファイルを文ごとに分割するジェネレータ．

    Args:
        f: 分割するファイル．

    Examples:

        >>> from rhoknp.units import Sentence
        >>> from rhoknp.utils.reader import chunk_by_sentence
        <BLANKLINE>
        >>> with open("example.knp") as f:
        ...     for knp in chunk_by_sentence(f):
        ...         sentence = Sentence.from_knp(knp)
    """
    buffer = []
    for line in f:
        if line.strip() == "":
            continue
        buffer.append(line)
        if line.strip() == Sentence.EOS_PAT:
            yield "".join(buffer)
            buffer = []
    if buffer:
        yield "".join(buffer)


def chunk_by_document(
    f: TextIO, doc_id_format: Union[Literal["default", "kwdlc", "wac"], Callable] = "default"
) -> Iterator[str]:
    """ファイルを文書ごとに分割するジェネレータ．

    Args:
        f: 分割するファイル．
        doc_id_format: 文書IDのフォーマット．

    Examples:

        >>> from rhoknp.units import Document
        >>> from rhoknp.utils.reader import chunk_by_document
        <BLANKLINE>
        >>> with open("example.knp") as f:
        ...     for knp in chunk_by_document():
        ...         document = Document.from_knp(knp)

    .. note::
        文書IDのフォーマットとして指定可能なのは以下の通り：
            * "default": 文ID (S-ID) の最後のハイフン以前を文書IDとみなす．
                (例) # S-ID:A-X-1 -> 文書ID: A-X
            * "kwdlc": KWDLCの文IDから文書IDを取り出す．
                (例) # S-ID:w201106-0000060050-1 -> 文書ID: w201106-0000060050
            * "wac": WACの文IDから文書IDを取り出す．
                (例) # S-ID:wiki00100176-00 -> 文書ID: wiki00100176

        関数が指定された場合，文解析結果の先頭行から文書IDを取り出す関数とみなす．
        例えば default 相当の処理を行うには以下のような関数を渡す．

            >>> def default_doc_id_format(line: str) -> str:
            ...     return line.lstrip("# S-ID:").split("-")[0]
    """
    if isinstance(doc_id_format, str):
        if doc_id_format == "default":
            extract_doc_id = _extract_doc_id(Sentence.SID_PAT)
        elif doc_id_format == "kwdlc":
            extract_doc_id = _extract_doc_id(Sentence.SID_PAT_KWDLC)
        elif doc_id_format == "wac":
            extract_doc_id = _extract_doc_id(Sentence.SID_PAT_WAC)
        else:
            raise ValueError(f"Invalid doc_id_format: {doc_id_format}")
    elif callable(doc_id_format):
        extract_doc_id = doc_id_format
    else:
        raise ValueError(f"Invalid doc_id_format: {doc_id_format}")

    prev_doc_id: Optional[str] = None
    buffer: List[str] = []
    for sentence in chunk_by_sentence(f):
        doc_id = extract_doc_id(sentence.split("\n")[0])
        if buffer and (prev_doc_id != doc_id or doc_id is None):
            yield "".join(buffer)
            buffer = []
        buffer.append(sentence)
        prev_doc_id = doc_id
    if buffer:
        yield "".join(buffer)


def _extract_doc_id(pat: re.Pattern) -> Callable[[str], Optional[str]]:
    """文書IDを抽出する関数を返す．

    Args:
        pat: 文書IDを抽出する正規表現．
    """

    def extract_doc_id(line: str) -> Optional[str]:
        if match_sid := re.match(r"# S-ID: ?(\S*)( .+)?$", line):
            sid_string = match_sid.group(1)
            match = pat.match(sid_string)
            if match is None:
                logger.warning(f"Invalid S-ID: {sid_string}")
                return None
            return match.group("did")
        return None

    return extract_doc_id
